@@ -7,7 +7,7 @@ use Carp 'croak','carp';
 eval 'use Net::Netmask';
 eval 'use Net::ISP::Balance::ConfigData';
 
-our $VERSION    = '1.11';
+our $VERSION    = '1.14';
 
 =head1 NAME
 
@@ -194,11 +194,12 @@ sub new {
 	verbose   => 0,
 	echo_only => 0,
 	services  => {},
-	rules_directory => $class->default_rules_directory,
-	lsm_conf_file   => $class->default_lsm_conf_file,
-	lsm_scripts_dir => $class->default_lsm_scripts_dir,
-	bal_conf_file   => $conf,
-	dummy_data      => $dummy_test_data,
+	rules_directory    => $class->default_rules_directory,
+	lsm_conf_file      => $class->default_lsm_conf_file,
+	lsm_scripts_dir    => $class->default_lsm_scripts_dir,
+	bal_conf_file      => $conf,
+	keep_custom_chains => 1,
+	dummy_data         => $dummy_test_data,
     },ref $class || $class;
 
     $self->_parse_configuration_file($conf);
@@ -364,6 +365,22 @@ sub echo_only {
     $d;
 }
 
+=head2 $boolean = $bal->keep_custom_chains([boolean]);
+
+Get/set the keep_custom_chains flag. If this is true (default), then
+any custom iptables chains, such as those created by miniunpnpd or
+fail2ban, will be restored after execution of the firewall rules. If
+false, then these rules were be flushed.
+
+=cut
+
+sub keep_custom_chains {
+    my $self = shift;
+    my $d    = $self->{keep_custom_chains};
+    $self->{keep_custom_chains} = shift if @_;
+    $d;
+}
+
 =head2 $result_code = $bal->sh(@args)
 
 Pass @args to the shell for execution. If echo_only() is set to true,
@@ -1041,6 +1058,29 @@ sub _service_field {
     $s->{$field};
 }
 
+sub _save_custom_chains {
+    my $self = shift;
+    for my $table ('filter','nat','mangle') {
+	my @rules = split("\n",`sudo iptables -t $table -S`);
+	# find custom chains
+	my $mine    = 'MARK-|REJECTPERM|DROPGEN|DROPINVAL|DROPPERM|DROPSPOOF|DROPFLOOD|DEBUG';
+	my @chains  = grep {!/^-N ($mine)/} grep {/^-N (\S+)/} @rules or next;
+	s/^-N // foreach @chains;
+	my $chains  = join '|',map {quotemeta($_)} @chains;
+	my @targets = grep {/-(?:j|A|I) (?:$chains)/} @rules;
+	$self->{_custom_chains}{$table} = [(map {"-N $_"} @chains),@targets];
+    }
+}
+
+sub _restore_custom_chains {
+    my $self = shift;
+    my $custom_chains = $self->{_custom_chains} or return;
+    for my $table (keys %{$custom_chains}) {
+	my @rules = @{$custom_chains->{$table}} or next;
+	$self->iptables([map {"-t $table $_"} @rules]);
+    }
+}
+
 =head1 FILES AND PATHS
 
 These are methods that determine where Net::ISP::Balance finds its
@@ -1310,6 +1350,14 @@ sub _collect_interfaces {
     my $self = shift;
     my $s    = $self->{svc_config} or return;
 
+    # OH HELL - HAVE TO CHANGE LOGIC TO DEAL WITH VIRTUAL INTERFACES
+    # 1. From _ip_addr_show get all the inet XXX.XXX.XXX.XXX lines and calculate
+    #    corresponding network and virtual interface.
+    # 2. Record mapping of network to virtual interface in a hash (%vin?)
+    # 3. When going through the routes, replace $dev with virtual interface name
+    # 4. In (keys %devs) loop, create an inner loop for each inet found and replace
+    #    device with correct virtual device.
+
     # get interfaces with assigned addresses
     my $a    = $self->_ip_addr_show;
     my (undef,@ifs)  = split /^\d+: /m,$a;
@@ -1320,52 +1368,70 @@ sub _collect_interfaces {
 	($dev,$config);
     } @ifs;
 
+    # find virtual interfaces
+    my (%vnet,%vif);
+    for my $dev (keys %ifs) {
+	my $info = $ifs{$dev};
+	while ($info =~ /inet (\d+\.\d+\.\d+\.\d+)(?:\/(\d+))?.+?(\S+)$/mg) {
+	    my ($addr,$bits,$vdev) =  ($1,$2,$3);
+	    $bits ||= 32;
+	    my ($peer)      = $info =~ /peer\s+(\d+\.\d+\.\d+\.\d+)/;
+	    my $block       = Net::Netmask->new2("$addr/$bits") 
+		or die "unable to derive address for $dev: $Net::Netmask::error\nifconfig = \n$info";
+	    $vnet{$dev}{"$block"}    = $vdev;
+	    $vif{$dev}{$vdev}{block} = $block;
+	    $vif{$dev}{$vdev}{addr}  = $addr;
+	}
+    }
+
     # get existing routes
     my (%gws,%nets);
     my $r    = $self->_ip_route_show;
     my @routes = split /^(?!\s)/m,$r;
     chomp(@routes);
     foreach (@routes) {
-	while (/(\S+)\s+via\s+(\S+)\s+dev\s+(\w+)/g) {
+	while (/(\S+)\s+via\s+(\S+)\s+dev\s+(\S+)/g) {
 	    my ($net,$gateway,$dev) = ($1,$2,$3);
 	    ($net) = /^(\S+)/ if $net eq 'nexthop';
-	    $nets{$dev} = $net unless $net eq 'default';
-	    $gws{$dev}  = $gateway;
+	    my $vdev  = $vnet{$dev}{$net} || $dev;
+	    $nets{$vdev} = $net unless $net eq 'default';
+	    $gws{$vdev}  = $gateway;
 	}
     }
                                
     # map devices to services
     my %devs;
     for my $svc (keys %$s) {
-	my $dev = $s->{$svc}{dev};
-	$devs{$dev}=$svc;
+	my $vdev = $s->{$svc}{dev};
+	$devs{$vdev}=$svc;
     }
 
     my %ifaces;
     my $counter = 0;
     for my $dev (sort keys %devs) {
 	my $info      = $ifs{$dev} or next;
-	my $svc       = $devs{$dev};
-	my $role      = $s->{$svc}{role};
 	my $running   = $info =~ /[<,]UP[,>]/;
-	my ($addr,$bits)= $info =~ /inet (\d+\.\d+\.\d+\.\d+)(?:\/(\d+))?/;
-	$bits ||= 32;
-	my ($peer)      = $info =~ /peer\s+(\d+\.\d+\.\d+\.\d+)/;
-	my $block       = Net::Netmask->new2("$addr/$bits") 
-	    or die "unable to derive address for $dev: $Net::Netmask::error\nifconfig = \n$info";
-	my $gw          = $gws{$dev}  || $peer                    || $self->_dhcp_gateway($dev) || $block->nth(1);
-	my $net         = $nets{$dev} || ($peer?"$peer/32":undef) || "$block";
-	$ifaces{$svc} = {
-	    dev     => $dev,
-	    running => $running,
-	    gw      => $gw,
-	    net     => $net,
-	    ip      => $addr,
-	    fwmark  => $role eq 'isp' ? ++$counter : undef,
-	    table   => $role eq 'isp' ?   $counter : undef,
-	    role    => $role,
-	    ping    => $s->{$svc}{ping},
-	    weight  => $s->{$svc}{weight},
+	my ($peer)    = $info =~ /peer\s+(\d+\.\d+\.\d+\.\d+)/;
+	for my $vdev (keys %{$vif{$dev}}) {
+	    my $addr  = $vif{$dev}{$vdev}{addr};
+	    my $block = $vif{$dev}{$vdev}{block};
+	    my $svc   = $devs{$vdev};
+	    my $role  = $s->{$svc}{role};
+	    my $net   = $nets{$dev} || ($peer?"$peer/32":undef) || "$block";
+	    my $gw    = $gws{$dev}  || $peer || $self->_dhcp_gateway($dev) || $block->nth(1);
+	    $ifaces{$svc} = {
+#		dev     => $vdev,
+		dev     => $dev,    # otherwise, iptables will croak!!!
+		running => $running,
+		gw      => $gw,
+		net     => $net,
+		ip      => $addr,
+		fwmark  => $role eq 'isp' ? ++$counter : undef,
+		table   => $role eq 'isp' ?   $counter : undef,
+		role    => $role,
+		ping    => $s->{$svc}{ping},
+		weight  => $s->{$svc}{weight},
+	    }
 	};
     }
     return \%ifaces;
@@ -1449,7 +1515,14 @@ needed to create the balancing firewall.
 
 sub set_firewall {
     my $self = shift;
-    $self->base_fw_rules();
+    if ($self->keep_custom_chains) {
+	$self->_save_custom_chains;
+	$self->base_fw_rules();
+	$self->_restore_custom_chains;
+    } else {
+	$self->base_fw_rules();
+    }
+
     $self->balancing_fw_rules();
     $self->sanity_fw_rules();
     $self->nat_fw_rules();
@@ -1725,6 +1798,7 @@ END
     # packets from LAN
     for my $lan ($self->lan_services) {
 	my $landev = $self->dev($lan);
+	my $src    = $self->net($lan);
 	
 	if (@up > 1) {
 	    print STDERR "# creating balanced mangling rules\n" if $self->verbose;
@@ -1733,7 +1807,7 @@ END
 	    for my $svc (sort {$probabilities->{$b} <=> $probabilities->{$a}} @up) {
 		my $table       = $self->mark_table($svc);
 		my $probability = $probabilities->{$svc};
-		$self->iptables("-t mangle -A PREROUTING -i $landev -m conntrack --ctstate NEW -m statistic --mode random --probability $probability -j $table");
+		$self->iptables("-t mangle -A PREROUTING -i $landev -s $src -m conntrack --ctstate NEW -m statistic --mode random --probability $probability -j $table");
 	    }
 	}
 
@@ -1741,18 +1815,19 @@ END
 	    my $svc = $up[0];
 	    print STDERR "# forcing all traffic through $svc\n" if $self->verbose;
 	    my $table  = $self->mark_table($svc);
-	    $self->iptables("-t mangle -A PREROUTING -i $landev -m conntrack --ctstate NEW -j $table");
+	    $self->iptables("-t mangle -A PREROUTING -i $landev -s $src -m conntrack --ctstate NEW -j $table");
 	}
 
-	$self->iptables("-t mangle -A PREROUTING -i $landev -m conntrack --ctstate ESTABLISHED,RELATED -j CONNMARK --restore-mark");
+	$self->iptables("-t mangle -A PREROUTING -i $landev -s $src -m conntrack --ctstate ESTABLISHED,RELATED -j CONNMARK --restore-mark");
     }
 
     # inbound packets from WAN
     for my $wan ($self->isp_services) {
 	my $dev   = $self->dev($wan);
 	my $table = $self->mark_table($wan);
-	$self->iptables("-t mangle -A PREROUTING -i $dev -m conntrack --ctstate NEW -j $table");
-	$self->iptables("-t mangle -A PREROUTING -i $dev -m conntrack --ctstate ESTABLISHED,RELATED -j CONNMARK --restore-mark");
+	my $src   = $self->net($wan);
+	$self->iptables("-t mangle -A PREROUTING -i $dev -s $src -m conntrack --ctstate NEW -j $table");
+	$self->iptables("-t mangle -A PREROUTING -i $dev -s $src -m conntrack --ctstate ESTABLISHED,RELATED -j CONNMARK --restore-mark");
     }
 
 }
